# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: 2021 The Elixir Team
# SPDX-FileCopyrightText: 2012 Plataformatec

defmodule Mix.Tasks.Compile.Protocols do
  @moduledoc false
  use Mix.Task

  # TODO: Deprecate me on Elixir v1.23
  def run(_args) do
    :noop
  end
end
